var wsServer = false;
var wssPort = 8888;

var socCount = 0;     /* try to detect browser reload verses exit */
var httpPort = 8080;
var lastTime = Date.now();
var chromeUI = null;  /* chrome UI process */
var rootDirs = [];    /* external root directories for web server */

var cancelExit = false;

const spawn = require("child_process").spawn;
const fs = require('fs');
const connect = require('connect');
const serverStatic = require('serve-static');
const app = connect();

main();

/*
 *  ======== usage ========
 */
function usage()
{
    console.log('Usage: node <polymerROV dir>/src/nodeserver.js [options]');
    console.log('Options:');
    console.log('    --noCloudAgent');
    console.log('        if set, WebSocket server is started and it spawns');
    console.log('        an xdc.rov.monserver');
    console.log('    --httpPort');
    console.log('        HTTP port for the local server, defaults to 8080');
    console.log('    --httpRoot');
    console.log('        add HTTP root directory for the local server to read.');
    console.log('        This option can be repeated to add multiple roots.');
    console.log('    --wsPort');
    console.log('        port for the WebSocket server, defaults to 8888');
    process.exit(0);
}

/*
 *  ======== main ========
 */
function main()
{
    if (process.argv.length < 2) {
        usage();
    }

    /* shift out argv[0] (the node command itself) */
    process.argv.shift();

    /* loop through all other arguments looking for options */
    do {
        var arg = process.argv.shift();
        if (arg == "--noCloudAgent") {
            wsServer = true;
        }
        else if (arg == "--httpPort") {
            httpPort = process.argv.shift();
        }
        else if (arg == "--httpRoot") {
            rootDirs.push(process.argv.shift());
        }
        else if (arg == "--wsPort") {
            wssPort = process.argv.shift();
        }
        else if (arg.charAt(0) != '-') {
            continue;
        }
        else {
            console.log("Invalid argument: " + arg);
            usage();
        }
    } while (process.argv.length > 0);

    if (httpPort == wssPort) {
        console.log("HTTP port and WebSocket port must not be same.");
        usage();
    }

    /* catch and handle unhandled exceptions */
    process.on('uncaughtException', function (er) {
       /* ignore "addr in use" exception, so we can start chrome using
        * existing nodeserver */
       if (er.message.indexOf("listen EADDRINUSE") == -1) {
           console.error(er.stack);
           process.exit(1);
       }
    });

    /* append any directories from ROV_XDC_HTTPROOTS to rootDirs */
    if (process.env.ROV_XDC_HTTPROOTS != null) {
        var roots = process.env.ROV_XDC_HTTPROOTS.split(';');
        rootDirs = rootDirs.concat(roots);
    }

    /* root directory for the HTTP server is <location_of_this_file>/.. */
    var idxDir = __dirname.replace(/\\/g, '/');         // for Windows (sigh)
    idxDir = idxDir.substr(0, idxDir.lastIndexOf('/')); // strip "/src"

    /* configure the HTTP server */
    app.use(httpMonitor); /* watch all requests to detect browser exit */

    rootDirs.splice(0, 0, idxDir); /* add idxDir as the first HTTP root */
    for (var i = 0; i < rootDirs.length; i++) {
        var root = rootDirs[i];
        if (root != null && root != "") {
            app.use(serverStatic(root));
        }
    }

    /* start the HTTP server */
    console.log('Starting web server from [' + rootDirs
                + '] on port ' + httpPort);
    app.listen(httpPort);

    /* optionally start a WebSocket monServer */
    if (wsServer) {
        startMonServer(idxDir);
    }

    /* start the browser */
    startChrome();
}

/*
 *  ======== startChrome ========
 */
function startChrome() 
{
    var chrome = findChrome();
    var params = ["--app=http://localhost:" + httpPort + "/local.html",
                  "--window-size=950,600"
                 ];
    trace("spawning chrome: " + chrome + " " + params);
    chromeUI = spawn(chrome, params, {detached: false, stdio: 'ignore'});

    chromeUI.on("error", 
        function(msg) {
            trace("ROV chromeUI spawn error: " + msg);
        });

    if (chromeUI.pid == 0) {
        chromeUI = null;
        trace("ROV chromeUI couldn't start!");
        process.exit(1);
    }

    chromeUI.on("exit", 
        function(code, signal) {
            trace("Chrome exited: " + code 
                  + (signal != null ? (", signal = " + signal) : ""));
        });
}

/*
 *  ======== startMonServer ========
 */
function startMonServer(idxDir)
{
    /* This server runs within Node.js, and spawns a Rhino instance that runs
     * xdc.rov.monserver.
     */
    var WSServer = require('ws').Server;

    var wss = new WSServer({port: wssPort});
    trace("WebSocket server started at port " + wssPort);

    wss.on('connection', function connection(socket) {
        var buffer = ""; /* needed to assemble chunks of one long view */

        socCount++;
        trace("Connected to browser socket: " + socCount);

        var xs = process.env.XDCROOT + "/xs";
        if (process.platform == "win32") {
            xs += ".exe";
        }

        /* ensure xs exists and is executable, for a good error message */
        fs.accessSync(xs, fs.X_OK);

        var params = ["--xdcpath=" + process.env.XDCPATH,
                      "xdc.rov.monserver"
        ];

        trace("spawning monserver: " + xs + " " + params);
        var monServer = spawn(xs, params, {
                         detached: false,
                         cwd:      idxDir + "/../monserver"});

        monServer.on("error", function(msg) {
            trace("ROV Server spawn error: " + msg);
        });
        if (monServer.pid == 0) {
            monServer = null;
            trace("ROV monServer couldn't start!");
            process.exit(1);
        }

        monServer.stdout.setEncoding("utf8");
        monServer.stderr.setEncoding("utf8");

        monServer.stdout.on("data", function(data) {
            trace("ROV Server stdout: " + data);

            /* append incoming data to any unprocessed data */
            buffer = buffer + data;

            /* process end-of-message */
            while (buffer.indexOf("<EOM>") != -1) {
                var first = buffer.indexOf("<BOM>");
                var last = buffer.indexOf("<EOM>");

                /* If there is <EOM> in the buffer, there should be <BOM>
                 * somewhere up front. If <BOM> is not there, or there is only
                 * one after <EOM> something went wrong, we'll dump the buffer
                 * and throw an error.
                 */
                if (first == -1 || first > last) {
                    console.log("nodeserver buffer content: " + buffer);
                    throw new Error("nodeserver buffer is in an invalid state");
                }

                /* Send content between <BOM><EOM> pair */
                socket.send(buffer.slice(first + 5, last));

                /* In case there is another <BOM> already in the buffer */
                if (buffer.indexOf("<BOM>", last) != -1) {
                    buffer = buffer.slice(buffer.indexOf("<BOM>", last));
                }
                else {
                    buffer = "";
                }
            }
        });

        monServer.stdout.on("close", function() {
            trace("ROV Server closed stdout stream");
        });

        monServer.stderr.on("data", function(data) {
            trace("ROV Server stderr: " + data);
        });

        monServer.stderr.on("close", function() {
            trace("ROV Server closed stderr stream");
        });

        socket.on('message', function incoming(msg) {
            var command = JSON.parse(msg);
            trace("Received from browser: " + command);
            if (monServer != null) {
                monServer.stdin.write(command);
            }
            else {
                trace("  command dropped: monserver null!");
            }
        });

        socket.on('close', function close() {
            trace("browser disconnected socket: " + socCount);
            if (monServer != null) {
                monServer.kill('SIGTERM');
                monServer = null;
            }

            /* decrement socket count */
            if (socCount > 0) {
                socCount--;
            }

            /* on last socket close, check whether we should exit */
            if (socCount == 0) {
                checkForExit();
            }
        });
    });
}

/*
 *  ======== trace ========
 */
function trace(msg) {
    var now = Date.now();
    console.log("nodeserver [ms " + (now - lastTime) + "]: " + msg);
    lastTime = now;
}

/*
 *  ======== checkForExit ========
 *  Exit standalone nodeserver when we don't see a load of the startup page
 */
function checkForExit() {
    trace("checking for exit ... (cancel = " + cancelExit + ")");
    if (cancelExit == true) {
        cancelExit = false;
        return;
    }
    trace("    exiting.");
    process.exit(0);
}

/*
 *  ======== httpMonitor ========
 *  Watch for loads of the "standalone" ROV2 startup page
 */
function httpMonitor(req, res, next){
    if (String(req.url).indexOf("/local.html") != -1) {
        trace("http req: '" + req.url + "', socCount = " + socCount);
    }

    /* prevent last socket close from terminating on reloads */
    cancelExit = true;
    setTimeout(function() {cancelExit = false}, 1000);
    
    next(); /* service next HTTP service function */
}

/*
 *  ======== findChrome ========
 */
function findChrome()
{
    var chrome = null;

    var paths = [
        "C:/Program Files (x86)/Google/Chrome/Application/chrome.exe",
        "C:/Program Files/Google/Chrome/Application/chrome.exe",
        process.env.LOCALAPPDATA + "/Google/Chrome/Application/chrome.exe"
    ];

    for (var i = 0; i < paths.length; i++) {
        try {
            fs.accessSync(paths[i], fs.X_OK);
            chrome = paths[i];
            break;
        } 
        catch (x) {
            ;
        }
        
    }
    if (chrome == null) {
        throw new Error("can't find chrome in this list: [" + paths + "]\n");
    }

    return (chrome);
}
